"""
workbench.py contains the class workbench(), which creates the technical setup. 
The class workbench() allows you to install and manage all the components 
necessary for the successfully productionizing your model.

The components you need are:

1. Docker: as a container technology https://www.docker.com
2. VirtualBox: as a driver https://www.virtualbox.org
3. Minikube: as a local Kubernetes cluster https://minikube.sigs.k8s.io/
4. Kubectl: as a Kubernetes CLI https://kubectl.docs.kubernetes.io

Slots:
--------
wd : str
    Stores the current working directory
current_status : str
    Stores the current status of the minikube cluster
platform : str
    Stores the platform the host is running on
py_version : string
    Stores the version of Python in use
dk_installed : boolean
    Stores if Docker is already installed
vb_installed : boolean
    Stores if VirtualBox is already installed
kc_installed : boolean
    Stores if kubectl is already installed
mk_installed : boolean
    Stores if minikube is already installed
dk_prev_installed : boolean
    Stores if Docker was already installed
vb_prev_installed : boolean
    Stores if VirtualBox was already installed
kc_prev_installed : boolean
    Stores if kubectl was already installed
mk_prev_installed : boolean
    Stores if minikube was already installed
current_projects : list
    Stores all active projects on the workbench
cpus_used : string
    Stores the CPUs allocated to the workbench
memory_used : string
    Stores the memory allocated to the workbench
"""

# import libs
import subprocess
import os
import sys
import re
import warnings

# setup the class
class workbench:

    # define the class object
    def __init__(self):

        # store the working directory
        self.wd = os.getcwd()

        # store the status
        self.current_status = 'initialized'

        # store the platform
        self.platform = sys.platform

        # check python version
        major_v = str(sys.version_info[0])
        minor_v = str(sys.version_info[1])
        micro_v = str(sys.version_info[2])

        # write to slot
        self.py_version = str(major_v + '.' + minor_v + '.' + micro_v)
        
        # store if components are installed
        self.vb_installed = None
        self.dk_installed = None
        self.kc_installed = None
        self.mk_installed = None

        # check if components were installed
        self.__check_installed()

        # store prev installed
        self.dk_prev_installed = self.dk_installed
        self.vb_prev_installed = self.vb_installed
        self.kc_prev_installed = self.kc_installed
        self.mk_prev_installed = self.mk_installed

        # store component versions
        self.dk_version = None
        self.vb_version = None
        self.kc_version = None
        self.mk_version = None

        # store memory usage
        self.cpus_used = None
        self.memory_used = None

        # welcome message
        welcome_message = """

                                                    Welcome to
        ██████╗ ██████╗  ██████╗ ██████╗ ██╗   ██╗ ██████╗████████╗██╗ ██████╗ ███╗   ██╗██╗███████╗███████╗
        ██╔══██╗██╔══██╗██╔═══██╗██╔══██╗██║   ██║██╔════╝╚══██╔══╝██║██╔═══██╗████╗  ██║██║╚══███╔╝██╔════╝
        ██████╔╝██████╔╝██║   ██║██║  ██║██║   ██║██║        ██║   ██║██║   ██║██╔██╗ ██║██║  ███╔╝ █████╗  
        ██╔═══╝ ██╔══██╗██║   ██║██║  ██║██║   ██║██║        ██║   ██║██║   ██║██║╚██╗██║██║ ███╔╝  ██╔══╝  
        ██║     ██║  ██║╚██████╔╝██████╔╝╚██████╔╝╚██████╗   ██║   ██║╚██████╔╝██║ ╚████║██║███████╗███████╗
        ╚═╝     ╚═╝  ╚═╝ ╚═════╝ ╚═════╝  ╚═════╝  ╚═════╝   ╚═╝   ╚═╝ ╚═════╝ ╚═╝  ╚═══╝╚═╝╚══════╝╚══════╝
                                                                                                    
                                                                                                    
                        productionize is an open-source lightweight ML deployment tool.
                        You can containerize, deploy and ship your model, without ever
                        having to leave your beloved Python.


        Before we start, we will first setup your local workbench. The workbench is a collection of tools
        we need to do all those amazing things. You are running the following setup:

        Platform: {operating_system}
        Python:   {python_version}
        Cores:    {number_cores}

        The main components of the workbench are containerization and orchestration tools. I know there are
        many different tools out there, but we are using:

                  Already on machine      Purpose
        ------------------------------------------------------
        Docker:         {dk_exists}        Containerization
        VirtualBox:     {vb_exists}        Cluster Hypervisor
        Kubectl:        {kc_exists}        Kubernetes CLI
        Minikube:       {mk_exists}        Local Kubernetes


        You can download the missing components and configure them using: workbench.setup(). Don't worry
        though, using workbench.delete(), you will be able to cleanly uninstall all the components again.

        """.format(operating_system = self.platform,
                   python_version = self.py_version,
                   number_cores = os.cpu_count(),
                   dk_exists = self.dk_prev_installed,
                   vb_exists = self.vb_prev_installed,
                   kc_exists = self.kc_prev_installed,
                   mk_exists = self.mk_prev_installed)

        # print welcome message
        print(welcome_message)

        # check if brew exists
        try:
            
            # update brew before starting
            subprocess.call('brew update'.split(), stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)

        # if it crashed
        except:

            # raise Exception
            raise Exception('homebrew is not installed on your machine, please install it here: https://brew.sh')

        # add project slot
        self.current_projects = []
        
    # helper function to check if components are installed
    def __check_installed(self):

        """
        Helper method to check if components are already installed.

        This function tests if the main components Docker, VirtualBox, Kubectl
        and Minikube are already installed.
        """

        # try to call docker
        try:

            # call docker
            subprocess.call('docker version'.split(), stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)

            # if it did not crash, it is installed
            self.dk_installed = True

        # if it crashes
        except:

            # it is not installed
            self.dk_installed = False
        
        # try to call virtualbox
        try:

            # call virtualbox
            subprocess.call('virtualbox --help'.split(), stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)

            # if it did not crash, it is installed
            self.vb_installed = True

        # if it crashes
        except:

            # it is not installed
            self.vb_installed = False

        # try to call kubectl
        try:

            # call kubectl
            subprocess.call('kubectl config view'.split(), stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)

            # if it did not crash, it is installed
            self.kc_installed = True

        # if it crashes
        except:

            # it is not installed
            self.kc_installed = False

        # try to call minikube
        try:

            # call minikube
            subprocess.call('minikube version'.split(), stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)

            # if it did not crash, it is installed
            self.mk_installed = True

        # if it crashes
        except:

            # it is not installed
            self.mk_installed = False

    # main function to debug
    def setup_debug(self, issue = "docker"):

        """
        main method to debug the setup() method.

        This function helps to debug any issues in the setup of the
        workbench. This method is supposed to give users some hints
        in case something goes wrong.

        Parameters
        ----------
        issue : string
            indicates what component should be investigated, the default is
            "docker", alternatives are "virtualbox", "kubectl" and "minikube".
        """

        # check if issue docker:
        if issue == 'docker':

            # install docker
            subprocess.call('brew cask install docker'.split())

        # check if issue virtualbox
        if issue == 'virtualbox':

            # install virtualbox
            subprocess.call('brew cask install virtualbox --force'.split())
        
        # check if issue kubectl
        if issue == 'kubectl':

            # install kubectl
            subprocess.call('brew install kubectl'.split())

        # check if issue minikube
        if issue == 'minikube':

            # install minikube
            subprocess.call('brew install minikube'.split())

            # link cli version
            subprocess.call('brew link --overwrite kubernetes-cli'.split())

    # helper function to install docker
    def __install_docker(self):

        """
        helper method to install Docker.

        This function downloads and sets up Docker Desktop. The user can then 
        use the Docker Desktop interface to log in.
        """

        # check if Docker is already installed
        if self.dk_prev_installed:

            #  collect version
            dk_version = subprocess.check_output('docker version --format "{{.Server.Version}}"'.split())
            dk_version = '.'.join(re.findall(r'\d+', str(dk_version)))

        # if it is not installed already
        else:

            # try to install docker
            try:

                # install docker
                dk_install_success = subprocess.call('brew cask install docker'.split(), stdout=subprocess.DEVNULL)

                # check if it worked
                if dk_install_success == 0:

                    # set dk_install_success to True
                    dk_install_success = True
                
                # if it didn't work
                else:

                    # set dk_install_success to False
                    dk_install_success = False

            # handle crash
            except:

                # set dk_install_success to False
                dk_install_success = False

            # check if it worked
            if dk_install_success:

                # print message
                print ('> Docker was successfully installed')

                # update version
                dk_version = 'log in first'

            # if it didn't work
            else:

                # stop function
                raise Exception('Docker could not be installed, run setup_debug(issue = "docker") to print detailed logs')

        # write docker version to self
        self.dk_version = dk_version
        
    # helper function to install virtualbox
    def __install_virtualbox(self):

        """
        helper method to install VirtualBox.

        This function downloads and sets up VirtualBox. The users can now use
        VirtualBox to start VMs on their local machine.
        """

        # check if Virtualbox is already installed
        if self.vb_prev_installed:

            #  collect version
            vb_version = subprocess.check_output('vboxmanage --version'.split())
            vb_version = '.'.join(re.findall(r'\d+', str(vb_version))[0:3])

        # if it is not installed already
        else:

            # try to install virtualbox
            try:

                # install virtualbox
                vb_install_success = subprocess.call('brew cask install virtualbox --force'.split(), stdout=subprocess.DEVNULL)

                # check if it worked
                if vb_install_success == 0:

                    # set vb_install_success to True
                    vb_install_success = True
                
                # if it didn't work
                else:

                    # set vb_install_success to False
                    vb_install_success = False

            # handle crash
            except:

                # set vb_install_success to False
                vb_install_success = False

            # check if it worked
            if vb_install_success:

                # try to collect version
                try:

                    # check the version
                    vb_version = subprocess.check_output('vboxmanage --version'.split())
                    vb_version = '.'.join(re.findall(r'\d+', str(vb_version))[0:3])

                    # print message
                    print ('> VirtualBox was successfully installed')
                
                # handle exception
                except:

                    # stop function
                    raise Exception('VirtualBox could not be installed, run setup_debug(issue = "virtualbox") to print detailed logs')

            # if it didn't work
            else:

                # stop function
                raise Exception('VirtualBox could not be installed, run setup_debug(issue = "virtualbox") to print detailed logs')

        # write VirtualBox version to self
        self.vb_version = vb_version

    # helper function to install kubectl
    def __install_kubectl(self):

        """
        helper method to install kubectl.

        This function downloads and sets up kubectl as a kubernetes-cli.
        """

        # check if kubectl is already installed
        if self.kc_prev_installed:

            #  collect version
            kc_version = subprocess.check_output('kubectl version --client=true'.split())
            kc_version = '.'.join(re.findall(r'\d+', str(kc_version))[0:2])

        # if it is not installed already
        else:

            # try to install kubectl
            try:

                # install kubectl
                kc_install_success = subprocess.call('brew install kubectl'.split(), stdout=subprocess.DEVNULL)

                # check if it worked
                if kc_install_success == 0:

                    # set kc_install_success to True
                    kc_install_success = True
                
                # if it didn't work
                else:

                    # set kc_install_success to False
                    kc_install_success = False

            # handle crash
            except:

                # set kc_install_success to False
                kc_install_success = False

            # check if it worked
            if kc_install_success:

                # try to catch version
                try:

                    # check the version
                    kc_version = subprocess.check_output('kubectl version --client=true'.split())
                    kc_version = '.'.join(re.findall(r'\d+', str(kc_version))[0:2])

                    # print message
                    print ('> Kubectl was successfully installed')

                # handle exception
                except:

                    # stop function
                    raise Exception('Kubectl could not be installed, run setup_debug(issue = "kubectl") to print detailed logs')

            # if it didn't work
            else:

                # stop function
                raise Exception('Kubectl could not be installed, run setup_debug(issue = "kubectl") to print detailed logs')

        # write Kubectl version to self
        self.kc_version = kc_version

    # helper function to install minikube
    def __install_minikube(self):

        """
        helper method to install Minikube.

        This function downloads and sets up Minikube, which is a local 
        Kubernetes implementation.
        """

        # check if minikube is already installed
        if self.mk_prev_installed:

            #  collect version
            mk_version = subprocess.check_output('minikube version'.split())
            mk_version = '.'.join(re.findall(r'\d+', str(mk_version))[0:3])

        # if it is not installed already
        else:

            # try to install minikube
            try:

                # install minikube
                mk_install_success = subprocess.call('brew install minikube'.split(), stdout=subprocess.DEVNULL)

                # link kubectl
                subprocess.call('brew link --overwrite kubernetes-cli'.split(), stdout=subprocess.DEVNULL)

                # check if it worked
                if mk_install_success == 0:

                    # set mk_install_success to True
                    mk_install_success = True
                
                # if it didn't work
                else:

                    # set mk_install_success to False
                    mk_install_success = False

            # handle crash
            except:

                # set kc_install_success to False
                mk_install_success = False

            # check if it worked
            if mk_install_success:

                # try to collect version
                try:

                    # check the version
                    mk_version = subprocess.check_output('minikube version'.split())
                    mk_version = '.'.join(re.findall(r'\d+', str(mk_version))[0:3])

                    # print message
                    print ('> Minikube was successfully installed')

                # handle exception
                except:

                    # stop function
                    raise Exception('Minikube could not be installed, run setup_debug(issue = "minikube") to print detailed logs')

            # if it didn't work
            else:

                # stop function
                raise Exception('Minikube could not be installed, run setup_debug(issue = "minikube") to print detailed logs')

        # write VirtualBox version to self
        self.mk_version = mk_version

    # main function to setup the workbench
    def setup(self, report = True):

        """
        main method to setup the workbench.

        This function downloads and sets up all components necessary for the
        workbench. The components include Docker, VirtualBox, Kubectl and
        Minikube.

        Parameters
        ----------
        report : boolean
            if True a process report is printed with the progress on the
            installation
        """

        # install docker
        self.__install_docker()

        # install virtualbox
        self.__install_virtualbox()

        # install kubectl
        self.__install_kubectl()

        # install minikube
        self.__install_minikube()

        # check if all components can be detected
        self.__check_installed()

        # check if the report should be printed
        if report:

            # build report
            report = """

            Setup Report:
            -------------

            This is an automatically generated report on the setup process of your local workbench. The set
            of components required for the workbench to work were processed as follows:

                                Status                     Version
            ---------------------------------------------------------------------
            Docker:         installed == {dk_exists}        {dk_version}
            VirtualBox:     installed == {vb_exists}        {vb_version}
            Kubectl:        installed == {kc_exists}        {kc_version}
            Minikube:       installed == {mk_exists}        {mk_version}

            In case not all components could be installed, please consult the setup_debug() method. This
            method prints out the system logs when installing and configuring the components.


            ! | If Docker was newly installed, you will need to log in to start Docker Desktop and log in
            ! | or create an account at Dockerhub https://hub.docker.com. In case you have any questions,
            ! | please consult: https://www.docker.com/products/docker-desktop.

            """.format(dk_exists = self.dk_version,
                       vb_exists = self.vb_prev_installed,
                       kc_exists = self.kc_prev_installed,
                       mk_exists = self.mk_prev_installed,
                       dk_version = self.dk_version,
                       vb_version = self.vb_version,
                       kc_version = self.kc_version,
                       mk_version = self.mk_version)

            # print report
            print (report)

        # update status
        self.current_status = 'installed'

    # main function to start cluster
    def start_cluster(self, cpus = '2', memory = '2G'):

        """
        main method to start the workbench cluster

        This function initiates and starts the local workbench cluster. The
        cluster is a Kubernetes based composition of Docker, VirtualBox,
        Kubectl and Minikube.

        Parameters
        ----------
        cpus : string
            indicates the number of cores the cluster can use as resources
        memory : string
            indicates the amount of memory the cluster can use
        """

        # try to start minikube
        try:

            # start minikube
            subprocess.call(str('minikube start --driver=virtualbox --cpus=' + cpus + ' --memory=' + memory).split())

            # update status
            self.current_status = 'running'

            # update resource usage
            self.cpus_used = cpus
            self.memory_used = memory

        # handle exception
        except:

            # update status
            self.current_status = 'crashed'

            # raise exception
            raise Exception('I could not start the cluster')

    # main function to open a new project
    def open_project(self, name = None):

        """
        main method to open projects on the workbench cluster.

        This function opens a new project on the workbench cluster. Technically
        speaking, this function creates a new Kubernetes namespace.

        Parameters
        ----------
        name : string
            indicates the project name, but should not have spaces or any
            separators, others than "-"
        """

        # check if project name was given
        if name is None:

            # assign default project name
            name = 'my-project'
        
        # if name was given
        else:

            # ensure the name complies with standards
            name = name.replace('_', '-').replace('/', '-')

        # try to create a new namespace
        try:

            # create new namespace
            subprocess.call(str('kubectl create namespace ' + name).split(), stdout=subprocess.DEVNULL)

            # send message
            print (str('> Successfully created project: ' + name))

            # add to list of projects
            self.current_projects.append(name)
        
        # if it crashed
        except:

            # raise Exception
            raise Exception('I could not create the new project')
    
    # main method to list all current projects
    def list_projects(self):

        """
        main method to list all projects on the workbench cluster.

        This function lists all projects that are currently on the workbench.
        """

        # print list with all projects on the workbench
        return self.current_projects

    # main method to list all products in project
    def list_products(self, project):

        """
        main method to list all products in a projec.

        This function lists all products that are currently deployed in a
        project on the workbench.

        Parameters
        ----------
        project : string
            String that gives the name of the project, for which products should be listed
        """

        # try to list all products
        try:

            # retrieve list of products
            command = str("kubectl get pods --template=\'{{range .items}}{{.metadata.name}}{{","}}{{end}}\' -n " + project)
            product_list = os.popen(command).read()

            # return product list
            return product_list

        # handle exception
        except:

            # print message
            print (str('It seems that there are no products deployed for your project: ' + project))

    # main method to delete a project
    def delete_project(self, name = None):

        """
        main method to delete projects on the workbench cluster.

        This function deletes a project on the workbench cluster. Technically
        speaking, this function deletes a Kubernetes namespace. To do so,
        the methods requires a project name

        Parameters
        ----------
        name : string
            indicates the project name, but should not have spaces or any
            separators, others than "-"
        """

        # check if name was given
        if name is None:

            # break function
            raise Exception('You need to provide a name for the project you want to delete')
        
        # if name was provided
        else:

            # try to delete the project
            try:

                # delete project    
                subprocess.call(str('kubectl delete namespaces ' + name), stdout=subprocess.DEVNULL)

                # print message
                print (str('> Successfully deleted project: ' + name))

            # handle exception
            except:

                # raise Exception
                raise Exception('I could not delete the project')

            # try to delete project out of list
            try:

                # delete project out of list
                self.current_projects.remove(name)
            
            # handle exception
            except:

                # print warning
                warnings.warn('I could not find the project in self.current_projects')

    # main function to give status report on cluster
    def inspect(self):

        """
        main method to build a status report on cluster.

        This function builds a report on the current workbench. It shows how
        many products, projects and resources are bound.
        """

        # check if all components are running
        self.__check_installed()

        # calculate number of projects
        project_counter = len(self.current_projects)

        # initialize product counter
        product_counter = 0

        # loop over all projects to count products
        for project in self.current_projects:

            # get the products deployed
            products_in_project = self.list_products(project = project)
            
            # split in list and count length
            product_counter = product_counter + len(products_in_project.split(','))

            # return
            return product_counter

        # build report
        report = """

        Workbench Report:
        -----------------

        This is an automatically generated report on your workbench. Your
        workbench consists of four main components:

                              Status                     Version
        ---------------------------------------------------------------------
        Docker:         installed == {dk_exists}        {dk_version}
        VirtualBox:     installed == {vb_exists}        {vb_version}
        Kubectl:        installed == {kc_exists}        {kc_version}
        Minikube:       installed == {mk_exists}        {mk_version}

        You are running this session with Python {python_version}. Your machine
        has {number_cores}. In total you allocated the following resources to
        the workbench:

        Cores:      {cpus_used}
        Memory:     {memory_used}

        Your workbench is currently hosting {product_counter} product in 
        {project_counter} project(s).

        """.format(operating_system = self.platform,
                   python_version = self.py_version,
                   number_cores = os.cpu_count(),
                   dk_exists = self.dk_installed,
                   vb_exists = self.vb_installed,
                   kc_exists = self.kc_installed,
                   mk_exists = self.mk_installed,
                   cpus_used = self.cpus_used,
                   memory_used = self.memory_used,
                   project_counter = project_counter,
                   product_counter = product_counter)

        # print report
        print (report)

    # main function to stop cluster
    def stop_cluster(self):

        """
        main method to stop the workbench cluster.

        This function stops the running workbench cluster.
        """

        # try to stop minikube
        try:

            # stop cluster
            subprocess.call('minikube stop'.split(), stdout=subprocess.DEVNULL)

            # update status
            self.current_status = 'stopped'

            # print message
            print ('> Successfully stopped the cluster')

        # handle exception
        except:

            # update status
            self.current_status = 'not responding'

            # raise exception
            raise Exception('I could not stop the cluster')

    # main function to uninstall workbench
    def uninstall(self, docker = None, kubectl = None, virtualbox = None, minikube = None, report = True):

        """
        main method to uninstall all components.

        This function cleanly uninstalls all components used for the workbench.
        The user can decide if a specific component is removed. If no input is
        given, the components that already were present, the components are 
        kept. For those components not present, the removal is processed.

        Parameters
        ----------
        docker : boolean
            indicates if docker should be removed from the machine
        kubectl : boolean
            indicates if kubectl should be removed from the machine
        virtualbox : boolean
            indicates if virtualbox should be removed from the machine
        minikube : boolean
            indicates if minikube should be removed from the machine
        report : True
            if True a process report is printed with the progress on the
            installation
        """

        # check if minikube should be deleted
        if minikube is None:

            # check if it was installed before
            if self.mk_prev_installed:

                # don't delete if it was there before
                minikube = False
            
            # if it wasn't installed before
            else:

                # delete if it wasn't there before
                minikube = True

        # check if kubectl should be deleted
        if kubectl is None:

            # check if it was installed before
            if self.kc_prev_installed:

                # don't delete if it was there before
                kubectl = False

            # if it wasn't installed before
            else:

                # delete if it wasn't there before
                kubectl = True

        # check if virtualbox should be deleted
        if virtualbox is None:

            # check if it was installed before
            if self.vb_prev_installed:

                # don't delete if it was there before
                virtualbox = False 

            # if it wasn't installed before
            else:

                # delete if it wasn't there before
                virtualbox = True
        
        # check if docker should be deleted
        if docker is None:

            # check if it was installed before
            if self.dk_prev_installed:

                # don't delete if it was there before
                docker = False

            # if it wasn't installed before
            else:

                # delete if it wasn't there before
                docker = True

        # check if Minikube should be deleted
        if minikube:

            # try to delete minikube
            try:

                # delete minikube
                mk_deleted = subprocess.call('brew uninstall minikube --force'.split(), stdout=subprocess.DEVNULL)

                # check if it worked
                if mk_deleted == 0:

                    # print message
                    print ('> Successfully deleted Minikube')

                # if it didn't work
                else:

                    # raise Exception
                    raise Exception('I could not delete Minikube')

            # handle exception
            except:

                # raise Exception
                raise Exception('I could not delete Minikube')

        # check if kubectl should be deleted
        if kubectl:

            # try to delete kubectl
            try:

                # delete kubectl
                kc_deleted = subprocess.call('brew uninstall kubectl --force'.split(), stdout=subprocess.DEVNULL)

                # check if it worked
                if kc_deleted == 0:

                    # print message
                    print ('> Successfully deleted Kubectl')
                
                # if it didn't work
                else:

                    # raise Exception
                    raise Exception('I could not delete Kubectl')

            # handle exception
            except:

                # raise Exception
                raise Exception('I could not delete Kubectl')

        # check if virtualbox should be deleted
        if virtualbox:

            # try to delete virtualbox
            try:

                # delete virtualbox
                vb_deleted = subprocess.call('brew cask uninstall virtualbox --force'.split(), stdout=subprocess.DEVNULL)

                # check if it worked
                if vb_deleted == 0:

                    # print message
                    print ('> Successfully deleted VirtualBox')

                # if it didn't work
                else:

                    # raise Exception
                    raise Exception('I could not delete VirtualBox')

            # handle exception
            except:

                # raise Exception
                raise Exception('I could not delete VirtualBox')

        # check if docker should be deleted
        if docker:

            # try to delete docker
            try:

                # delete docker
                dk_deleted = subprocess.call('brew cask zap docker --force'.split(), stdout=subprocess.DEVNULL)
                dk_deleted = subprocess.call('brew cask zap docker --force'.split(), stdout=subprocess.DEVNULL)

                # check if it worked
                if dk_deleted == 0:

                    # print message
                    print ('> Successfully deleted Docker')
                
                # if it didn't work
                else:

                    # raise Exception
                    raise Exception('I could not delete Docker')

            # handle exception
            except:

                # raise Exception
                raise Exception('I could not delete Docker')

        # check if they are still installed
        self.__check_installed()

        # check if report should be printed
        if report:

            # build report
            report = """

            Deletion Report:
            ----------------

            This is an automatically generated report on the uninstall process of your local workbench. The set
            of components required for the workbench to work were processed as follows:

                        Still on machine               
            --------------------------------
            Docker:         {dk_exists}       
            VirtualBox:     {vb_exists}  
            Kubectl:        {kc_exists}
            Minikube:       {mk_exists}


            """.format(dk_exists = self.dk_installed,
                    vb_exists = self.vb_installed,
                    kc_exists = self.kc_installed,
                    mk_exists = self.mk_installed)

            # print report
            print (report)


# load libs
from setuptools import setup
import productionize

# read in README.md
with open("README.md", "r") as fh:
    long_description = fh.read()

# catch the version
current_version = productionize.__version__

# define the setup
setup(name='productionize',
      version=current_version,
      description='Lightweight ML Deployment Platform',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/LJStroemsdoerfer/productionize',
      author='Lukas Jan Stroemsdoerfer',
      author_email='ljstroemsdoerfer@gmail.com',
      license='MIT',
      packages=['productionize'],
      zip_safe=False)


"""
product.py contains the class product(), which is at the center of productionizing 
your ML model. The product() class takes an API script and allows you to easily
containerize the API. You can then deploy the container, export the image or call
the deployed API.

Slots:
--------
wd : string
    Current working directory
current_status : string
    Current status of the product
"""

# import libs
import subprocess
import os
import sys

# setup the class
class product:

    # describe the class
    def __init__(self, project = None):

        # store working directory
        self.wd = os.getcwd()

        # store status of product
        self.current_status = 'initialized'

        # check python version
        major_v = str(sys.version_info[0])
        minor_v = str(sys.version_info[1])
        micro_v = str(sys.version_info[2])

        # write to slot
        self.py_version = str(major_v + '.' + minor_v + '.' + micro_v)

        # check if project name was given
        if project is None:

            # give default name
            project = 'my-project'

        # store project name
        self.project_name = project

        # store path to api file
        self.api_file = None

        # store path to requirements file
        self.requirements_file = None

        # store port to deploy
        self.port = None

        # store product name
        self.product_name = None

        # store service url
        self.service_url = None
    
        # build report
        report = """

        Product Report:
        ---------------

        This is an automatically generated report on the status of your product 
        deployment. productionize considers a product to be a deployable API.

          Your Product
        -----------------------
        Name:      
        Project:    {project}
        Status:     {status}

        With the product class you can easily deploy your python API, without
        leaving python. The product class is capable of building and deploying 
        Docker containers on your workbench. In addition, the class provides 
        out-of-the-box functionalities to extract and push your container to
        any registry you want.

        """.format(project = self.project_name,
                   status = self.current_status)

        # print report
        print (report)

    # helper method to build Dockerfile
    def __build_dockerfile(self):

        """
        Private method to build a Dockerfile.
        This function first builds a Docker file from a python script and 
        a requirements.txt.
        """

        # try create a dockerfile
        try:

            # store dk_file_path
            dk_file_path = str(self.wd + '/Dockerfile')

            # open a new file
            subprocess.call(str('touch ' + dk_file_path).split(), stdout=subprocess.DEVNULL)

            # write content
            content = """\
            FROM python:{version}
            RUN mkdir -p /api
            COPY {api_file} /api/api.py
            COPY {requirements_file} /api/requirements.txt
            RUN python -m pip install -r /api/requirements.txt
            EXPOSE {port}
            ENTRYPOINT ["python", "api/api.py"]\
            """.format(version=self.py_version,
                    api_file = self.api_file,
                    requirements_file = self.requirements_file,
                    port = int(self.port))

            # open Dockerfile
            file = open(dk_file_path, "w")

            # write content to Dockerfile
            file.write(content)

            # close connection
            file.close()

            # write file path to self
            self.dk_file_path = dk_file_path

            # check if it worked
            if self.dk_file_path is None:

                # raise Exception
                raise Exception('I could not create the Dockerfile in your current working directory: ' + self.wd)

        # handle exception
        except:

            # raise exception
            raise Exception(str('I could not create the Dockerfile in your current working directory: ' + self.wd))

    # main function to deploy API
    def prepare_deployment(self, api_file, requirements_file, port, name = None):

        """
        Main method to prepare the deployment.

        This function builds a dockerfile out of the script and the require-
        ments file provided. Together with the system python version, the
        container is defined. After this function, the user can take a look
        to the Dockerfile and even adjust it, however, at the users own risk.

        Parameters
        ----------
        api_file : string
            String with the path to the python api file
        requirements_file : string
            String with the path to the requirements file
        port : string
            String with the port number to expose
        name : string
            String with the name of your deployment
        """

        # check the api file
        if isinstance(api_file, str):

            # try to read in first lines
            try:

                # read in file
                with open(api_file) as file:
                    first_line_api = file.readline()

                # store to self, but ensure there is no tilde
                self.api_file = api_file.replace('~','')

            # exception handling
            except:

                # raise Exception
                raise Exception(str('I could not find your file: ' + api_file))
        
        # if it is not a string
        else:

            # raise Exception
            raise Exception('api_file arg should be a path anf a filename to your python file that defines the api: e.g. your_folder/your_script.py')
    
        # check the requirements file
        if isinstance(requirements_file, str):

            # try to read in first lines
            try:

                # read in file
                with open(requirements_file) as file:
                    first_line_req = file.readline()

                # store to self, but ensure there is no tilde
                self.requirements_file = requirements_file.replace('~', '')

            # exception handling
            except:

                # raise Exception
                raise Exception(str('I could not find your file: ' + requirements_file))
        
        # if it is not a string
        else:

            # raise Exception
            raise Exception('requirements_file arg should be a path anf a filename to your requirements.txt: e.g. your_folder/your_reqs.txt')

        # check if port is string
        if isinstance(port, str):

            # store to self
            self.port = str(port)

        # if it is not a string
        else:

            # raise exception
            raise Exception('port arg should be a string with the desired exposing port: e.g. port = "8000"')

        # check if name was provided
        if name is None:

            # give default name
            name = 'my-product'
        
        # store name in self
        self.product_name = name.replace('_', '-').replace('/', '-')

        # build Dockerfile
        self.__build_dockerfile()

        # change status
        self.current_status = 'ready to deploy'

        # build report
        report = """

        Deployment Report:
        ------------------

        This is an automatically generated report on the preparation of your
        deployment. The Dockerfile was built using the following files:

        {api_file}:
        ----------------
        {first_line_api}
        ----------------

        and

        {requirements_file}:
        ----------------
        {first_line_req}
        ----------------

        The Dockerfile is ready to be used. You can inspect your Dockerfile: 
        
        {dk_file_path}

        You can of course edit the file, however, this is at your own risk. The
        processes in the container currently run as root. You can change that
        of course, as long as your API allows you to.

          Your Product
        -----------------------
        Name:       {name}
        Project:    {project}
        Status:     {status}


        """.format(api_file = self.api_file,
                   first_line_api = first_line_api,
                   requirements_file = self.requirements_file,
                   first_line_req = first_line_req,
                   dk_file_path = self.dk_file_path,
                   name = self.product_name,
                   project = self.project_name,
                   status = self.current_status)
    
        # print report
        print (report)

    # helper method to create Dockerfile
    def __build_image(self):
        """
        Private method to build a Docker image.

        This function takes the Dockerfile and creates an image on the Minikube
        internal registry.
        """

        # try to create the image on the minikube registry
        try:

            # make this is run from wd
            os.chdir(self.wd)

            # build image from Dockerfile
            command = str('eval $(minikube -p minikube docker-env) && docker build -t ' + self.product_name + '-image:latest .')
            os.system(command)

        # handle exception
        except:

            # raise exception
            raise Exception('I could not build the Docker image from the Dockerfile. In case you edited the file, please check if that was correct.')

    # helper method to run a deployment
    def __run_deployment(self):
        """
        Private method to run a deployment.

        This function uses kubectl to run a deployment on minikube.
        """

        # try to run deployment
        try:
            
            # run deployment
            command = str('kubectl run ' + self.product_name + " --image=kubipy-image:latest --image-pull-policy='Never' -n " + self.project_name)
            os.system(command)

        # handle exception
        except:

            # raise exception
            raise Exception('I could not run the deployment from the Docker image, make sure the Dockerfile is working.')

    # helper method to expose pod
    def __expose_pod(self):

        """
        Private method to expose a deployment.

        This function exposes the pod that was just deployed.
        """

        # try to expose the pod
        try:

            # expose the pod
            command = str('kubectl expose pod ' + self.product_name + ' --port=' + self.port + ' --type=NodePort -n ' + self.project_name)
            subprocess.call(command.split(), stdout=subprocess.DEVNULL)

        # handle exception
        except:

            # raise exception
            raise Exception('I could not expose the service, make sure your workbench is setup properly')

    # helper method to get the url
    def __get_url(self):

        """
        Private method to get the url of a deployment.

        This function exposes the service on a minikube level an retreives
        the url.
        """

        # try to expose service
        try:

            # expose the service on minikube
            command = str('minikube service ' + self.product_name + '-n ' + self.project_name + ' --url')
            service_url = subprocess.check_output(command.split())

            # decode url
            service_url = str(service_url.decode("utf-8")).replace("\n", "")
                
            # add route warning
            service_url = service_url + str('/<your_route>')

            # write the url to self
            self.service_url = service_url

        # handle exception
        except:

            # raise exception
            raise Exception('I could not expose the service to your host machine. Make sure the workbench is properly setup.')

    # helper function to check if deployment exists
    def __check_pods(self):

        """
        Private method to check if a pod exists.

        This function checks, if a pod already exists on Minikube.
        """

        # try to check if service exists
        try:

            # check for service
            command = str('kubectl get pod ' + self.product_name + ' -n' + self.project_name)
            exists = subprocess.call(command.split(), stdout=subprocess.DEVNULL)

            # check result
            if exists == 0:

                # return False
                return True
            
            # if not 0, then False
            else:

                # return False
                return False
        
        # if it breaks, it doesn't exist
        except:

            # return False
            return False

    # main method to delete products
    def delete_deployment(self, product, project):

        """
        Main method to delete deployments.

        This function deletes the deployment of specific products and all
        Minikube artifacts with it.

        Parameters
        ----------
        product : string
            String that gives the name of the product deployment that should be deleted
        project : string
            String that gives the name of the project in which the product should be deleted
        """

        # try to delete the product deployment
        try:

            # delete the service
            command = str('kubectl delete service ' + product + ' -n ' + project)
            subprocess.call(command.split(), stdout=subprocess.DEVNULL)

            # delete the pod
            command = str('kubectl delete pod ' + product + ' -n ' + project)
            subprocess.call(command.split(), stdout=subprocess.DEVNULL)

        # handle exception
        except:

            # raise exception
            raise Exception(str('I could not delete the deployment of your product: ' + product))

    # main method to deploy product
    def deploy(self):

        """
        Main method to deploy the product.

        This function takes the Dockerfile and deploys it to the workbench.
        """

        # check if product is already prepared
        if self.dk_file_path is None:

            # raise Exception
            raise Exception('You first need to run prepare_deployment() before deploying your product.')

        # build docker image on Minikube registry
        self.__build_image()

        # check if already exists
        exists_already = self.__check_pods()

        # if it already exists delete it
        if exists_already:

            # delete product 
            self.delete_deployment(product = self.product_name,
                                   project = self.project_name)

        # run deployment
        self.__run_deployment()

        # expose deployment
        self.__expose_pod()

        # get url
        self.__get_url()

        # change the status
        self.current_status = 'deployed and healthy'

        # build report
        report = """

        Deployment Report:
        ------------------

        This is an automatically generated report on the status of your deployed
        product. Your API is now containerized and hosted on the workbench. You
        can access the API using:

        {service_url}

        You can call the API in whatever way it is designed. If you want to get
        rid of it, just use the delete_deployment() method. If you just want to
        update the API, you can just use prepare_deployment() to create a new
        Dockerfile and then deploy() again.

                  Your Product
        -----------------------
        Name:       {name}
        Project:    {project}
        Status:     {status}
        Access:     {service_url}

        If you want to export the image to your local machine just use the
        export_product() method. If you want to push it to another registry,
        you can use the push_product() method.
        """.format(service_url = self.service_url,
                   name = self.product_name,
                   project = self.project_name,
                   status = self.current_status)

        # print report
        print (report)